/*******************************************************************************
 * $Id: CompactGappedAlignment.h,v 1.12 2004/04/19 23:10:50 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef __CompactGappedAlignment_h__
#define __CompactGappedAlignment_h__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libGenome/gnDebug.h"
#include "libGenome/gnFilter.h"
#include "libGenome/gnSequence.h"
#include "libMems/SparseAbstractMatch.h"
#include "libMems/AbstractGappedAlignment.h"
#include "libMems/UngappedLocalAlignment.h"

#include "boost/pool/pool_alloc.hpp"
#include <algorithm>

#ifdef WIN32
#include "windows.h"
#endif

namespace mems {

/**
 * The CompactGappedAlignment stores a gapped alignment as a bit-vector
 * Rather than using one byte per aligned position, this class uses one bit, making
 * particularly space efficient
 */
class CompactGappedAlignment : public AbstractGappedAlignment< SparseAbstractMatch<> >
{
public:
	CompactGappedAlignment() : AbstractGappedAlignment< SparseAbstractMatch<> >(){};
	CompactGappedAlignment( uint seq_count, gnSeqI align_length );
	CompactGappedAlignment( std::vector< std::vector<bool> >& aln_mat );
	CompactGappedAlignment( std::vector< bitset_t >& aln_mat, gnSeqI alignment_length );
	
	template< class MatchType >
	CompactGappedAlignment( MatchType& m );

	CompactGappedAlignment* Clone() const { return new CompactGappedAlignment( *this ); }
	CompactGappedAlignment* Copy() const;
	virtual void Free();
	
	void SetAlignment( const std::vector< std::string >& seq_align );

	void SetAlignment( std::vector< bitset_t >& seq_align );

	// Inherited methods from AbstractMatch:
	virtual void Invert();
	virtual void CropStart(gnSeqI crop_amount);
	virtual void CropEnd(gnSeqI crop_amount);

	virtual void CropLeft(gnSeqI crop_amount, uint seqI);
	virtual void CropRight(gnSeqI crop_amount, uint seqI);

	void GetAlignment( std::vector< std::vector<bool> >& align_matrix ) const;

	void GetAlignment( std::vector< bitset_t >& align_matrix ) const;

	/** allows a peek at the data inside this alignment.  don't change it or the CompactGappedAlignment will become corrupt */
	const std::vector< bitset_t >& GetAlignment() const{ return align_matrix; }

//	friend void GetAlignment( const CompactGappedAlignment& ga, const std::vector< genome::gnSequence* >& seq_table, std::vector<std::string>& alignment );
	
	void GetColumn( gnSeqI col, std::vector<gnSeqI>& pos, std::vector<bool>& column ) const;

	/** returns true if the given row,column of the alignment has a gap character */
	virtual bool IsGap( uint seq, gnSeqI col ) const;
	/** translate a cga to a new coordinate system */
	void translate( CompactGappedAlignment& cga, uint cga_seq, uint my_seq );

	bool validate() const;
	bool validate_bitcount() const;

	void copyRange( CompactGappedAlignment& dest, gnSeqI left_column, gnSeqI length );

protected:
	// assume that the STL has a space-efficient vector<bool> specialization
//	std::vector< std::vector<bool> > align_matrix;	/**< aligned positions have true values, gaps are false */
	std::vector< bitset_t > align_matrix;
	std::vector< std::vector< size_t > > bcount;

	void create_bitcount();
	gnSeqI SeqPosToColumn( gnSeqI pos, const bitset_t& bvec, const std::vector< size_t >& index ) const;
};

static bool debug_cga = false;

inline
CompactGappedAlignment* CompactGappedAlignment::Copy() const
{
#ifndef _USE_BOOST_MATCH_ALLOCATOR
	return new CompactGappedAlignment(*this);
#else
	boost::fast_pool_allocator< CompactGappedAlignment > fpa;
	CompactGappedAlignment* newb = boost::fast_pool_allocator< CompactGappedAlignment >::allocate();
	fpa.construct(newb, *this);
	return newb;
#endif
}
inline
void CompactGappedAlignment::Free()
{
#ifndef _USE_BOOST_MATCH_ALLOCATOR
	delete this;
#else
	boost::fast_pool_allocator< CompactGappedAlignment > fpa;
	fpa.destroy(this);
	boost::fast_pool_allocator< CompactGappedAlignment >::deallocate(this);
#endif
}

inline
bool CompactGappedAlignment::validate() const
{
	if( !debug_cga )
		return true;
	bool good = true;
	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
		gnSeqI count = align_matrix[seqI].count();
		if( count > 0 && LeftEnd(seqI) == 0 )
		{
			good = false;
			std::cerr << "boner_McHoserknob\n";
			genome::breakHere();
		}
		if( (count == 0 || Length(seqI) == 0) && LeftEnd(seqI) != 0 )
		{
			good = false;
			std::cerr << "Length(" << seqI << "): " << Length(seqI) << std::endl;
			std::cerr << "LeftEnd(seqI): " << LeftEnd(seqI) << std::endl;
			std::cerr << "spumante explosion\n";
			genome::breakHere();
		}
		if( count != Length(seqI) )
		{
			std::cerr << "seqI: " << seqI << " count: " << count << "  Length(seqI): " << Length(seqI) << std::endl;
			std::cerr << "LeftEnd(seqI): " << LeftEnd(seqI) << std::endl;
			std::cerr << "lendo mismatcho\n";
			genome::breakHere();
			return false;
		}
//		std::vector< std::vector< size_t > > tmp_bcount = bcount;
//		create_bitcount();
//		if( !tmp_bcount == bcount )
//		{
//			good = false;
//			std::cerr << "bcount mismatch!!!\n";
//		}
//		bcount = tmp_bcount;

	}
	return 	validate_bitcount() && good;
}


inline
CompactGappedAlignment::CompactGappedAlignment( std::vector< bitset_t >& aln_mat, gnSeqI alignment_length ) :
AbstractGappedAlignment< SparseAbstractMatch<> >( aln_mat.size(), alignment_length ),
align_matrix( aln_mat ),
bcount( std::vector< std::vector< size_t > >( aln_mat.size() ) )
{
	create_bitcount();
	validate_bitcount();
}

inline
CompactGappedAlignment::CompactGappedAlignment( std::vector< std::vector<bool> >& aln_mat ) :
AbstractGappedAlignment< SparseAbstractMatch<> >( aln_mat.size(), aln_mat[0].size() ),
align_matrix( std::vector< bitset_t >( SeqCount(), bitset_t( aln_mat[0].size(), false ) ) ),
bcount( std::vector< std::vector< size_t > >( SeqCount() ) )
{
	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
		bitset_t& bvec = align_matrix[seqI];
		std::vector< bool >& vbool = aln_mat[seqI];
		size_t charI = 0;
		for( ; charI < vbool.size(); charI++ )
			if( vbool[charI] )
				bvec.set(charI);
	}
	create_bitcount();
	validate_bitcount();
}

template< class MatchType >
CompactGappedAlignment::CompactGappedAlignment( MatchType& m ) : 
AbstractGappedAlignment< SparseAbstractMatch<> >( m.SeqCount(), m.AlignmentLength() ),
bcount( std::vector< std::vector< size_t > >( m.SeqCount() ) )
{
	m.GetAlignment(align_matrix);

	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
		SetStart(seqI, m.Start(seqI));
		if( m.Start(seqI) != NO_MATCH )
			SetLength(m.Length(seqI), seqI);
		else
			SetLength(0, seqI);
	}

	create_bitcount();

	if( !validate() )
	{
		std::cerr << "kahnstruct error\n";
//		std::cerr << *this << endl;
//		std::cerr << m << endl;
	}
}

inline
void CompactGappedAlignment::SetAlignment( std::vector< bitset_t >& seq_align )
{
	swap( align_matrix, seq_align );
	seq_align.clear();
	if( align_matrix.size() > 0 )
		SetAlignmentLength( align_matrix[0].size() );
	else
		SetAlignmentLength(0);
	bcount = std::vector< std::vector<size_t> >(align_matrix.size());
	create_bitcount();
	validate_bitcount();
}

inline
void CompactGappedAlignment::GetAlignment( std::vector< bitset_t >& align_matrix ) const
{
	align_matrix = this->align_matrix;
}

inline
void CompactGappedAlignment::GetAlignment( std::vector< std::vector<bool> >& align_matrix ) const
{
	align_matrix = std::vector< std::vector<bool> >( SeqCount(), std::vector<bool>(AlignmentLength(), false) );
	for( uint seqI = 0; seqI < align_matrix.size(); seqI++ )
	{
		if( LeftEnd(seqI) == NO_MATCH )
			continue;
		size_t bmid = this->align_matrix[seqI].find_first();		
		for( gnSeqI ct = 0;;  )
		{
			align_matrix[seqI][bmid] = true;
			if( ++ct < Length(seqI) )
				bmid = this->align_matrix[seqI].find_next( bmid );
			else
				break;
		}
	}
}

inline
bool CompactGappedAlignment::IsGap( uint seq, gnSeqI col ) const
{
	return !align_matrix[seq][col];
}

static const unsigned INDEX_INTERVAL = 512;

inline
void CompactGappedAlignment::create_bitcount()
{
	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
//		if( LeftEnd(seqI) == NO_MATCH )
//			continue;
		bitset_t& bvec = align_matrix[seqI];
		bcount[seqI].clear();
		bcount[seqI].push_back(0);
		for( size_t indie = 0; indie + INDEX_INTERVAL <= bvec.size(); indie += INDEX_INTERVAL )
		{
			size_t end = indie + INDEX_INTERVAL;
			size_t ct = 0;
			for( size_t i = indie; i < end; ++i )
				ct += bvec.test(i);
			bcount[seqI].push_back( ct + bcount[seqI].back() );
		}
	}
}

inline
bool CompactGappedAlignment::validate_bitcount() const
{
	if( !debug_cga )
		return true;
	bool valid = true;	// innocent until proven guilty
	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
		gnSeqI count = align_matrix[seqI].count();
		size_t bc_len = align_matrix[seqI].size() / INDEX_INTERVAL;
		if( count < INDEX_INTERVAL && bcount[seqI].size() == 0 )
			continue;	// a-ok here
		if( bc_len + 1 != bcount[seqI].size() && (bcount[seqI].back() % INDEX_INTERVAL != 0) )
		{
			std::cerr << "bitcount problem, bc_len + 1: " << bc_len + 1 << " and bcount[seqI].size(): " << bcount[seqI].size() << std::endl;
			std::cerr << "count: " << count << " and bcount[seqI].back(): " << bcount[seqI].back() << std::endl;
			valid = false;
		}
		if( count - bcount[seqI].back() > INDEX_INTERVAL )
		{
			std::cerr << "bitcount problem, count: " << count << " and bcount[seqI].back(): " << bcount[seqI].back() << std::endl;
			valid = false;
		}
	}
	return valid;
}

inline 
gnSeqI CompactGappedAlignment::SeqPosToColumn( gnSeqI pos, const bitset_t& bvec, const std::vector< size_t >& index ) const
{
	std::vector<size_t>::const_iterator iter = std::lower_bound(index.begin(), index.end(), pos);
	--iter;
	size_t cur_pos = *iter;
	size_t col = iter - index.begin();
	col *= INDEX_INTERVAL;
	if( col == 0 )
		col = bvec.find_first();
	else
		col = bvec.find_next(col-1);
	for( ++cur_pos; cur_pos < pos; ++cur_pos )
		col = bvec.find_next(col);
	return col;
}

inline
void CompactGappedAlignment::translate( CompactGappedAlignment& cga, uint cga_seq, uint my_seq ) // const
{
	AbstractMatch::orientation my_orient = Orientation(my_seq);
//	if( my_orient == AbstractMatch::reverse )
//		Invert();	// TODO: remove dependency on this horrible hack and make the function const again

	if( cga.Length(cga_seq)  > Length(my_seq) )
	{
		std::cerr << "Oh scheisskopf.  What are you trying to do to me??\n";
		std::cerr << "cga.Length(" << cga_seq << "): " << cga.Length(cga_seq) << std::endl;
		std::cerr << "Length(" << my_seq << "): " << Length(my_seq) << std::endl;
		genome::breakHere();
	}

	gnSeqI prev_lend = cga.LeftEnd(cga_seq);
	gnSeqI prev_len = cga.Length(cga_seq);
	gnSeqI my_lend = LeftEnd(my_seq);
	gnSeqI my_len = Length(my_seq);
	gnSeqI my_count = 0;
	uint seqI = 0;

	// what assumptions should be made about cga?
	// does it already have the correct left-end relative to this?
	// no, it needs to have a left-end relative to the first aligned char in this
	size_t cur_bit = 0;

	// determine left_bit
	size_t left_bit = SeqPosToColumn(cga.LeftEnd(cga_seq), align_matrix[my_seq], bcount[my_seq]);
	// determine right_bit
	size_t right_bit = SeqPosToColumn(cga.RightEnd(cga_seq), align_matrix[my_seq], bcount[my_seq]);
	if( right_bit > 4000000000u )
	{
		std::cerr << "cga doesn't fit\n";
		std::cerr << "cga.RightEnd(cga_seq) " << cga.RightEnd(cga_seq) << std::endl;
		std::cerr << "RightEnd(my_seq): " << RightEnd(my_seq) << std::endl;
		std::cerr << "cga.LeftEnd(cga_seq) " << cga.LeftEnd(cga_seq) << std::endl;
		std::cerr << "LeftEnd(my_seq): " << LeftEnd(my_seq) << std::endl;
		std::cerr << "cga.AlignmentLength(): " << cga.AlignmentLength() << std::endl;
		std::cerr << "AlignmentLength(): " << AlignmentLength() << std::endl;
		genome::breakHere();
	}
	right_bit++;
	if( right_bit == 0 )
		right_bit = AlignmentLength();

	cga.SetLeftEnd(cga_seq,left_bit+1);

	// add on length of unaligned left and right sides
	size_t cga_left = cga.align_matrix[cga_seq].find_first();

	size_t somesize = (right_bit - left_bit) - cga.Length(cga_seq) + cga.AlignmentLength();

	size_t cga_bit = cga_left;
	size_t my_bit = left_bit;
	size_t xlat_bit = cga_left;
	size_t added_bits = 0;
	// copy in everything up to cga_left
	std::vector< bitset_t > xrated( cga.SeqCount(), bitset_t( somesize, false ) );
	for( size_t seqI = 0; seqI < xrated.size(); ++seqI )
		for( size_t asdf = cga.align_matrix[seqI].find_first(); asdf < cga_left; asdf = cga.align_matrix[seqI].find_next(asdf) )
			xrated[seqI].set(asdf);
	
	while(xlat_bit < somesize)
	{
		// copy the column in cga
		for( size_t seqI = 0; seqI < xrated.size(); ++seqI )
			xrated[seqI].set( xlat_bit, cga.align_matrix[seqI].test(cga_bit) );

		++cga_bit;
		++xlat_bit;

		if( xlat_bit >= somesize )
			break;

		if( cga.align_matrix[cga_seq][cga_bit] )
		{
			size_t next_bit = align_matrix[my_seq].find_next(my_bit);
			if( next_bit > 4000000000u )
				genome::breakHere();
			size_t diff = next_bit - my_bit;
			if( diff > 1 )
			{
				if( xlat_bit + diff - 1 >= somesize )
				{
					std::cerr << "ERRRORRR porker!!\n";
					genome::breakHere();
				}
				for( size_t i = xlat_bit; i < xlat_bit + diff - 1; ++i )
					xrated[cga_seq].set(i);
//				xrated[cga_seq].set_range(xlat_bit, xlat_bit + diff - 1);
				added_bits += diff-1;
			}
			my_bit = next_bit;
			xlat_bit += diff - 1;
		}
	}

	cga.align_matrix = xrated;
	cga.create_bitcount();
	cga.SetLength(cga.Length(cga_seq)+added_bits,cga_seq);
	cga.SetAlignmentLength(somesize);
	if( !cga.validate() )
	{
		std::cerr << "prev_lend: " << prev_lend << std::endl;
		std::cerr << "prev_len: " << prev_len << std::endl;
		std::cerr << "translate error\n";
		genome::breakHere();
	}
//	if( my_orient == AbstractMatch::reverse )
//		Invert();
}


inline
void CompactGappedAlignment::Invert(){
	for(uint seqI = 0; seqI < SeqCount(); seqI++)
	{
		if( LeftEnd(seqI) == NO_MATCH )
			continue;
		bitset_t& fwd = align_matrix[seqI];
		bitset_t rev(AlignmentLength());
		size_t r = AlignmentLength();
		for( size_t i = 0; i < fwd.size(); ++i )
			rev.set( --r, fwd.test(i) );
		fwd.swap(rev);
	}
	create_bitcount();
	AbstractGappedAlignment< SparseAbstractMatch<> >::Invert();
	if( !validate() )
	{
		std::cerr << "invert error\n";
	}
}

inline
void CompactGappedAlignment::CropStart(gnSeqI crop_amount){
	if( crop_amount > AlignmentLength() )
		Throw_gnEx( genome::SeqIndexOutOfBounds() );
	if( crop_amount == 0 )
		return;

	gnSeqI pre_alignlen = AlignmentLength();
	gnSeqI pre_lend0 = LeftEnd(0);

	std::vector<gnSeqI> pos;
	std::vector<bool> column;
	GetColumn( crop_amount-1, pos, column );

	for( uint i=0; i < SeqCount(); i++ ){
		if( LeftEnd(i) == NO_MATCH )
		{
			align_matrix[i].resize(AlignmentLength()-crop_amount);
			continue;
		}

		align_matrix[i] >>= crop_amount;	// why not shift left?  is this a bug in boost::dynamic_bitset?
		align_matrix[i].resize(AlignmentLength()-crop_amount);
		size_t char_count = Orientation(i) == AbstractMatch::forward ? pos[i] - LeftEnd(i) + 1 : RightEnd(i) - pos[i] + 1;

		if( pos[i] > 0 && char_count > 0 )
		{
			SetLength(Length(i)-char_count, i);
			if( Length(i) == 0 )
				SetStart(i, NO_MATCH);
			if( Orientation(i) == AbstractMatch::forward )
				SetStart(i, Start(i) + char_count);
		}else if( pos[i] == 0 && Orientation(i) == AbstractMatch::reverse )
		{
			// this sequence was completely obliterated by the crop
			SetLength(0, i);
			SetStart(i, NO_MATCH);
		}
	}

	SetAlignmentLength( AlignmentLength() - crop_amount );
	create_bitcount();
	if( !validate() )
	{
		std::cerr << "pre_lend0: " << pre_lend0 << std::endl;
		std::cerr << "pre_alignlen: " << pre_alignlen << std::endl;
		std::cerr << "CropStart error\n";
	}
}

inline
void CompactGappedAlignment::CropEnd(gnSeqI crop_amount){
	if( crop_amount > AlignmentLength() )
		Throw_gnEx( genome::SeqIndexOutOfBounds() );
	if( crop_amount == 0 )
		return;

	std::vector<gnSeqI> pos;
	std::vector<bool> column;
	GetColumn( AlignmentLength()-crop_amount, pos, column );

	for( uint i=0; i < SeqCount(); i++ ){
		align_matrix[i].resize( AlignmentLength() - crop_amount );
		if( LeftEnd(i) == NO_MATCH )
			continue;
		AbstractMatch::orientation orient = Orientation(i);
		if( pos[i] > 0 )
		{
			gnSeqI char_count = pos[i] - (orient == AbstractMatch::forward ? (column[i] ? 1 : 0 ) : (column[i] ? 0 : 1 ) );
			char_count = orient == AbstractMatch::forward ? char_count - LeftEnd(i) + 1 : RightEnd(i) - char_count;
			if( char_count == 0 && align_matrix[i].count() > 0)
			{
				std::cerr << "orienatation: " << (orient == AbstractMatch::forward ? "forward\n" : (orient == AbstractMatch::reverse ? "reverse\n" : "undef\n"));
				std::cerr << "lend: " << LeftEnd(i) << std::endl;
				std::cerr << "length: " << Length(i) << std::endl;
				std::cerr << "count: " << align_matrix[i].count() << std::endl;
			}
			gnSeqI deleted = Length(i) - char_count;
			SetLength(char_count, i);
			if( Length(i) == 0 )
				SetStart(i, 0);
			if( Start(i) < 0 )
				SetStart(i, Start(i)-deleted);
		}else if( orient == AbstractMatch::forward ){
			SetLength(0, i);
			SetStart(i, 0);
		}
	}
	SetAlignmentLength( AlignmentLength() - crop_amount );
	create_bitcount();
	if( !validate() )
		std::cerr << "CropEnd error\n";
}

inline
void CompactGappedAlignment::CropLeft(gnSeqI crop_amount, uint seqI)
{
	if( crop_amount == 0 )
		return;

	gnSeqI pre_len = Length(seqI);
	// count "crop_amount" characters into seqI and crop there
	if( Orientation(seqI) == AbstractMatch::forward )
	{
		size_t left_col = SeqPosToColumn(crop_amount, align_matrix[seqI], bcount[seqI]) + 1;
		CropStart(left_col);
	}else{
		size_t left_col = SeqPosToColumn(Length(seqI) - crop_amount + 1, align_matrix[seqI], bcount[seqI]);
		if( left_col > 4000000000u )
		{
			std::cerr << LeftEnd(seqI) << std::endl;
			std::cerr << LeftEnd(0) << std::endl;
			std::cerr << "bogus cropper cga\n";
		}
		CropEnd(AlignmentLength()-left_col);
	}
	if( Length(seqI) != pre_len - crop_amount )
	{
		std::cerr << LeftEnd(seqI) << std::endl;
		std::cerr << LeftEnd(0) << std::endl;
		std::cerr << "bad cropperLeftie\n";
	}
	if( !validate() )
		std::cerr << "CropLeft error\n";
}

inline
void CompactGappedAlignment::CropRight(gnSeqI crop_amount, uint seqI)
{
	if( crop_amount == 0 )
		return;

	gnSeqI pre_len = Length(seqI);
	gnSeqI pre_lend = LeftEnd(seqI);
	gnSeqI pre_lend0 = LeftEnd(0);
	if( Orientation(seqI) == AbstractMatch::forward )
	{
		// count "crop_amount" characters into seqI and crop there
		size_t right_col = SeqPosToColumn(Length(seqI) - crop_amount + 1, align_matrix[seqI], bcount[seqI]);
		CropEnd( AlignmentLength()-right_col );
	}else
	{
		size_t right_col = SeqPosToColumn(crop_amount, align_matrix[seqI], bcount[seqI]) + 1;
		if( right_col > 4000000000u )
		{
			std::cerr << LeftEnd(seqI) << std::endl;
			std::cerr << LeftEnd(0) << std::endl;
			std::cerr << "bogus cropper cga\n";
		}
		CropStart( right_col );
	}
	if( Length(seqI) != pre_len - crop_amount )
	{
		std::cerr << LeftEnd(seqI) << std::endl;
		std::cerr << LeftEnd(0) << std::endl;
		std::cerr << "bad cropperight\n";
	}
	if( !validate() )
		std::cerr << "CropRight error\n";
}

inline
void CompactGappedAlignment::GetColumn( gnSeqI col, std::vector<gnSeqI>& pos, std::vector<bool>& column ) const
{
	pos = std::vector<gnSeqI>(SeqCount(), NO_MATCH);
	column = std::vector<bool>(SeqCount(), false);
	for( uint seqI = 0; seqI < SeqCount(); seqI++ )
	{
		if( align_matrix[seqI][col] )
			column[seqI] = true;

		gnSeqI count = 0;
		if( LeftEnd(seqI) != NO_MATCH )
		{
			size_t col_index = col / INDEX_INTERVAL;
			for( size_t i = col_index * INDEX_INTERVAL; i <= col; i++ )
				count += align_matrix[seqI].test(i);
			count += bcount[seqI][col_index];
		}

		if( count > 0 && Orientation(seqI) == forward )
			pos[seqI] = LeftEnd(seqI) + count - 1;
		else if( Orientation(seqI) == reverse && !(count == Length(seqI) && !column[seqI]) )
			pos[seqI] = RightEnd(seqI) - count + 1;
	}
}

inline
void CompactGappedAlignment::copyRange( CompactGappedAlignment& dest, gnSeqI left_column, gnSeqI length )
{
	if( left_column + length > AlignmentLength() )
		Throw_gnEx( genome::SeqIndexOutOfBounds() );
//	if( length == 0 )
//		return;

	// first copy the coordinates
	dest = CompactGappedAlignment(SeqCount(), length);
	for( uint i=0; i < SeqCount(); i++ ){
		dest.SetStart(i, Start(i));
		if( Orientation(i) != AbstractMatch::undefined )
			dest.SetLength(Length(i), i);
	}
	// then trim the coordinates appropriately

	gnSeqI pre_alignlen = AlignmentLength();
	gnSeqI pre_lend0 = LeftEnd(0);

	std::vector< bitset_t > dest_mat(2, bitset_t(length));
	std::vector<gnSeqI> pos;
	std::vector<bool> column;
	std::vector<gnSeqI> left_cc(SeqCount(), 0);
	if( left_column > 0 )
	{
		GetColumn( left_column-1, pos, column );
		for( uint i=0; i < SeqCount(); i++ ){
			if( LeftEnd(i) == NO_MATCH )
				continue;

			size_t char_count = Orientation(i) == AbstractMatch::forward ? pos[i] - LeftEnd(i) + 1 : RightEnd(i) - pos[i] + 1;
			if( pos[i] > 0 && char_count > 0 )
			{
				left_cc[i] = char_count;
				dest.SetLength(Length(i)-char_count, i);
				if( dest.Length(i) == 0 )
					dest.SetStart(i, NO_MATCH);
				if( dest.Orientation(i) == AbstractMatch::forward )
					dest.SetStart(i, dest.Start(i) + char_count);
			}else if( pos[i] == 0 && dest.Orientation(i) == AbstractMatch::reverse )
			{
				// this sequence was completely obliterated by the crop
				dest.SetLength(0, i);
				dest.SetStart(i, NO_MATCH);
			}
		}
	}

// now trim up the right side...
	gnSeqI right_trim = AlignmentLength() - left_column - length;

	if( right_trim > 0 )
	{
		GetColumn( AlignmentLength()-right_trim, pos, column );

		for( uint i=0; i < SeqCount(); i++ ){
//			align_matrix[i].resize( AlignmentLength() - crop_amount );
			if( LeftEnd(i) == NO_MATCH )
				continue;
			AbstractMatch::orientation orient = Orientation(i);
			if( pos[i] > 0 )
			{
				gnSeqI char_count = pos[i] - (orient == AbstractMatch::forward ? (column[i] ? 1 : 0 ) : (column[i] ? 0 : 1 ) );
				char_count = orient == AbstractMatch::forward ? char_count - LeftEnd(i) + 1 : RightEnd(i) - char_count;
				char_count -= left_cc[i];
/*
				if( char_count == 0 && (align_matrix[i].count()-left_cc[i]-length) > 0)
				{
					std::cerr << "orienatation: " << (orient == AbstractMatch::forward ? "forward\n" : (orient == AbstractMatch::reverse ? "reverse\n" : "undef\n"));
					std::cerr << "lend: " << LeftEnd(i) << std::endl;
					std::cerr << "length: " << Length(i) << std::endl;
					std::cerr << "count: " << align_matrix[i].count() << std::endl;
					genome::breakHere();
				}
*/
				gnSeqI deleted = Length(i) - char_count;
				dest.SetLength(char_count, i);
				if( dest.Length(i) == 0 )
					dest.SetStart(i, 0);
				if( dest.Start(i) < 0 )
					dest.SetStart(i, dest.Start(i)-deleted);
			}else if( orient == AbstractMatch::forward ){
				dest.SetLength(0, i);
				dest.SetStart(i, 0);
			}
		}
	}

	for( size_t i = 0; i < dest_mat.size(); ++i )
		for( size_t j = 0; j < length; ++j )
			dest_mat[i][j] = align_matrix[i][j+left_column];
	dest.SetAlignment(dest_mat);

	dest.create_bitcount();
	if( !dest.validate() )
	{
		std::cerr << "pre_lend0: " << pre_lend0 << std::endl;
		std::cerr << "pre_alignlen: " << pre_alignlen << std::endl;
		std::cerr << "CropStart error\n";
	}

}

}

#endif // __CompactGappedAlignment_h__

